Attribute VB_Name = "Fonts"
Option Explicit

'==================================================================
'
' Font Enumeration and Fullname Subroutines
'
' 29 September 2001
'
' You're free in using this software, as long as the
' orignal copyright/author information remains intact.
' In other words, respecting the person(s) who made it
' possible for you to view/learn something of this software.
'
' Copyright  2001 - Diederik van der Boor
' http://www.codingdomain.com
'
'==================================================================


'------------------------------------------------------------------
' Enumeration variables; updated within callback

Private EnumWhich As FontTypes
Private EnumResult As FontEnumResult
Private EnumFontNum As Long


'------------------------------------------------------------------
' Font API's

Public Type FontData
  CharSet As FontCharset
  ClipPrecision As Boolean
  Escapement As Boolean
  Face As String
  FullName As String
  Height As Boolean
  Italic As Boolean
  Orientation As Boolean
  OutPrecision As Boolean
  PitchAndFamily As Boolean
  Quality As Boolean
  Script As String
  StrikeOut As Boolean
  Style As String
  Type As FontTypes
  Underline As Boolean
  Weight As Boolean
  Width As Boolean
End Type

' The only way to make a collection of structures
Public Type FontEnumResult
  Fonts() As FontData
  FontsFound As Boolean
End Type

Public Enum FontTypes
  RASTER_FONTTYPE = 1
  DEVICE_FONTTYPE = 2
  TRUETYPE_FONTTYPE = 4
End Enum

Public Enum FontCharset
  ANSI_CHARSET = 0
  DEFAULT_CHARSET = 1
  SYMBOL_CHARSET = 2
  SHIFTJIS_CHARSET = 128
  HANGEUL_CHARSET = 129
  GB2312_CHARSET = 134
  CHINESEBIG5_CHARSET = 136
  OEM_CHARSET = 255
  JOHAB_CHARSET = 130
  HEBREW_CHARSET = 177
  ARABIC_CHARSET = 178
  GREEK_CHARSET = 161
  TURKISH_CHARSET = 162
  THAI_CHARSET = 222
  EASTEUROPE_CHARSET = 238
  RUSSIAN_CHARSET = 204
  
  MAC_CHARSET = 77
  BALTIC_CHARSET = 186
End Enum



' Win32 SDK recommends the use of EnumFontFamiliesEx rather than the other versions:

Private Const LF_FACESIZE = 32
Private Const LF_FULLFACESIZE = 64

Private Type LOGFONT
  lfHeight As Long
  lfWidth As Long
  lfEscapement As Long
  lfOrientation As Long
  lfWeight As Long
  lfItalic As Byte
  lfUnderline As Byte
  lfStrikeOut As Byte
  lfCharSet As Byte
  lfOutPrecision As Byte
  lfClipPrecision As Byte
  lfQuality As Byte
  lfPitchAndFamily As Byte
  lfFaceName(LF_FACESIZE - 1) As Byte
End Type

Private Type ENUMLOGFONTEX
  elfLogFont As LOGFONT
  elfFullName(LF_FULLFACESIZE - 1) As Byte
  elfStyle(LF_FACESIZE - 1) As Byte
  elfScript(LF_FACESIZE - 1) As Byte
End Type

Private Type NEWTEXTMETRIC
  tmHeight As Long
  tmAscent As Long
  tmDescent As Long
  tmInternalLeading As Long
  tmExternalLeading As Long
  tmAveCharWidth As Long
  tmMaxCharWidth As Long
  tmWeight As Long
  tmOverhang As Long
  tmDigitizedAspectX As Long
  tmDigitizedAspectY As Long
  tmFirstChar As Byte
  tmLastChar As Byte
  tmDefaultChar As Byte
  tmBreakChar As Byte
  tmItalic As Byte
  tmUnderlined As Byte
  tmStruckOut As Byte
  tmPitchAndFamily As Byte
  tmCharSet As Byte
  ' Additional to TEXTMETRIC
  ntmFlags As Long
  ntmSizeEM As Long
  ntmCellHeight As Long
  ntmAveWidth As Long
End Type

Private Type FONTSIGNATURE
  fsUsb(4) As Long
  fsCsb(2) As Long
End Type

Private Type TEXTMETRIC
  tmHeight As Long
  tmAscent As Long
  tmDescent As Long
  tmInternalLeading As Long
  tmExternalLeading As Long
  tmAveCharWidth As Long
  tmMaxCharWidth As Long
  tmWeight As Long
  tmOverhang As Long
  tmDigitizedAspectX As Long
  tmDigitizedAspectY As Long
  tmFirstChar As Byte
  tmLastChar As Byte
  tmDefaultChar As Byte
  tmBreakChar As Byte
  tmItalic As Byte
  tmUnderlined As Byte
  tmStruckOut As Byte
  tmPitchAndFamily As Byte
  tmCharSet As Byte
End Type

Private Type NEWTEXTMETRICEX
  ntmTm As NEWTEXTMETRIC
  ntmFontSig As FONTSIGNATURE
End Type

' Declares:
Private Declare Function EnumFontFamiliesEx Lib "gdi32" Alias "EnumFontFamiliesExA" (ByVal hDc As Long, lpLogFont As LOGFONT, ByVal lpEnumFontProc As Long, ByVal lParam As Long, ByVal dw As Long) As Long
Private Declare Function CreateScalableFontResource Lib "gdi32" Alias "CreateScalableFontResourceA" (ByVal fHidden As Long, ByVal lpszResourceFile As String, ByVal lpszFontFile As String, ByVal lpszCurrentPath As String) As Long
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDest As Any, pSource As Any, ByVal ByteLen As Long)


' Str is unicode
Private Const IS_TEXT_UNICODE_ASCII16 = &H1
Private Const IS_TEXT_UNICODE_REVERSE_ASCII16 = &H10
Private Const IS_TEXT_UNICODE_STATISTICS = &H2
Private Const IS_TEXT_UNICODE_REVERSE_STATISTICS = &H20
Private Const IS_TEXT_UNICODE_CONTROLS = &H4
Private Const IS_TEXT_UNICODE_REVERSE_CONTROLS = &H40
Private Const IS_TEXT_UNICODE_SIGNATURE = &H8
Private Const IS_TEXT_UNICODE_REVERSE_SIGNATURE = &H80
Private Const IS_TEXT_UNICODE_ILLEGAL_CHARS = &H100
Private Const IS_TEXT_UNICODE_ODD_LENGTH = &H200
Private Const IS_TEXT_UNICODE_DBCS_LEADBYTE = &H400
Private Const IS_TEXT_UNICODE_NULL_BYTES = &H1000
Private Const IS_TEXT_UNICODE_UNICODE_MASK = &HF
Private Const IS_TEXT_UNICODE_REVERSE_MASK = &HF0
Private Const IS_TEXT_UNICODE_NOT_UNICODE_MASK = &HF00
Private Const IS_TEXT_UNICODE_NOT_ASCII_MASK = &HF000
Private Declare Function IsTextUnicode Lib "advapi32" (lpBuffer As Any, ByVal cb As Long, lpi As Long) As Long





'------------------------------------------------------------------
' Enumerate fonts from hDc.
' Optional parameters are limitations for the enumeration

Public Function FontEnumAll(ByVal hDc As Long, Optional Which As FontTypes = 0, Optional FaceName As String, Optional CharSet As FontCharset = ANSI_CHARSET) As FontEnumResult
  Dim LogFnt As LOGFONT
  
  ' Convert the face name into a byte array:
  Call CByteArray(FaceName, LogFnt.lfFaceName)
  LogFnt.lfCharSet = CharSet
  EnumWhich = CLng(Which)
  EnumFontNum = 0
  EnumWhich = Which
  EnumResult.FontsFound = False
  ReDim EnumResult.Fonts(0) As FontData
  
  ' Start the enumeration:
  If EnumFontFamiliesEx(hDc, LogFnt, AddressOf EnumFontCallback, 0, 0) = 0 Then
    Err.Raise 12345, "Enum Fonts", "Could not enumerate fonts!"
  Else
    FontEnumAll = EnumResult
  End If
End Function


'------------------------------------------------------------------
' This routine is called by the EnumFontFamiliesEx API,
' each time when a font name is found, because we
' told that API the memory address of this subroutine.

Private Function EnumFontCallback(ByVal lpelfe As Long, ByVal lpntme As Long, ByVal iFontType As Long, ByVal lParam As Long) As Long
  ' The callback function for EnumFontFamiliesEx.
  
  ' lpelf points to an ENUMLOGFONTEX structure, lpntme points to either
  ' a NEWTEXTMETRICEX (if true type) or a TEXTMETRIC (if non-true type)
  ' structure.
  
  Dim tLFEx As ENUMLOGFONTEX
  Dim fntData As FontData
  
  ' Get the ENUMLOGFONTEX structure
  Call CopyMemory(tLFEx, ByVal lpelfe, LenB(tLFEx))
  
  If EnumWhich = iFontType Or EnumWhich = 0 Then
    fntData.CharSet = tLFEx.elfLogFont.lfCharSet
    fntData.ClipPrecision = tLFEx.elfLogFont.lfClipPrecision
    fntData.Escapement = tLFEx.elfLogFont.lfEscapement
    fntData.Face = CStrFromBytes(tLFEx.elfLogFont.lfFaceName)
    fntData.FullName = CStrFromBytes(tLFEx.elfFullName)
    fntData.Height = tLFEx.elfLogFont.lfHeight
    fntData.Italic = tLFEx.elfLogFont.lfItalic
    fntData.Orientation = tLFEx.elfLogFont.lfOrientation
    fntData.OutPrecision = tLFEx.elfLogFont.lfOutPrecision
    fntData.PitchAndFamily = tLFEx.elfLogFont.lfPitchAndFamily
    fntData.Quality = tLFEx.elfLogFont.lfQuality
    fntData.Script = CStrFromBytes(tLFEx.elfScript)
    fntData.StrikeOut = tLFEx.elfLogFont.lfStrikeOut
    fntData.Style = CStrFromBytes(tLFEx.elfStyle)
    fntData.Type = iFontType
    fntData.Underline = tLFEx.elfLogFont.lfUnderline
    fntData.Weight = tLFEx.elfLogFont.lfWeight
    fntData.Width = tLFEx.elfLogFont.lfWidth

    ReDim Preserve EnumResult.Fonts(0 To EnumFontNum) As FontData
    EnumResult.FontsFound = True
    EnumResult.Fonts(EnumFontNum) = fntData
    EnumFontNum = EnumFontNum + 1
  End If
  
  ' Ask for more fonts:
  EnumFontCallback = 1
End Function



'------------------------------------------------------------------
' Other Funny subroutine, which extracts
' the full name of a font from a .TTF file


Public Property Get FontFullName(TrueTypeFontFile As String) As String
  Dim hFile As Integer
  Dim Buffer As String
  Dim FontName As String
  Dim TempName As String
  Dim Pos As Integer
  
  'Build name for new resource file in a temporary file, and call the API.
  TempName = App.Path & "\~FONT.TMP"
  
  If CreateScalableFontResource(1, TempName, TrueTypeFontFile, vbNullString) Then
    'The name sits behind the text "FONTRES:"
    hFile = FreeFile
    Open TempName For Binary Access Read As hFile
      Buffer = Space(LOF(hFile))
      Get hFile, , Buffer
      Pos = InStr(1, Buffer, "FONTRES:") + 8
      FontName = Mid(Buffer, Pos, InStr(Pos, Buffer, vbNullChar) - Pos)
      Buffer = ""
    Close hFile
    Kill TempName
  End If
  
  'Return the font name
  FontFullName = FontName
End Property



'------------------------------------------------------------------
' Converts an BStr (used in VB) into a char-array (used in C)

Private Sub CByteArray(Text As String, ByRef Dest() As Byte)
  Dim I As Integer
  
  If Len(Text) Then
    If UBound(Dest) - LBound(Dest) < Len(Text) Then
      Err.Raise 12345, "Byte Array", "Buffer Overrun!"
    ElseIf StrIsUnicode(Text) Then
      Err.Raise 12345, "Byte Array", "String is already unicode!"
    Else
      For I = 1 To Len(Text)
        Dest(I - 1) = Asc(Mid(Text, I, 1))
      Next I
      Dest(Len(Text)) = Asc(vbNullChar)
    End If
  End If
End Sub

Private Function CStrFromBytes(ByteArray) As String
  CStrFromBytes = StrConv(ByteArray, vbUnicode)
  CStrFromBytes = CStrFromStrZ(CStrFromBytes)
End Function


'------------------------------------------------------------------
' Is String Unicode?

Private Function StrIsUnicode(sBuffer As String) As Boolean
  ' Returns True if sBuffer evaluates to a Unicode string
  Dim dwRtnFlags As Long
  dwRtnFlags = IS_TEXT_UNICODE_UNICODE_MASK
  StrIsUnicode = IsTextUnicode(ByVal sBuffer, Len(sBuffer), dwRtnFlags)
End Function

'------------------------------------------------------------------
' Convert String From lpszStrings (Long-Pointer Zero-Terminated) Strings
Private Function CStrFromStrZ(szStr As String) As String
  ' Returns string before first null char encountered (if any)
  ' from either an ANSII or Unicode string buffer.
  
  If StrIsUnicode(szStr) Then szStr = StrConv(szStr, vbFromUnicode)
  If InStr(szStr, vbNullChar) Then
    CStrFromStrZ = Left(szStr, InStr(szStr, vbNullChar) - 1)
  Else
    CStrFromStrZ = szStr
  End If
End Function

